@@ -9,52 +9,154 @@
 import time
 import ctypes
 import copy
+import functools
+import warnings
 from math import log10
 
 from . import ca
 from . import dbr
 from .utils import is_string
 
+try:
+    from types import SimpleNamespace as Namespace
+except ImportError:
+    from argparse import Namespace
+
 _PVcache_ = {}
 
-def get_pv(pvname, form='time',  connect=False,
-           context=None, timeout=5.0, **kws):
-    """get PV from PV cache or create one if needed.
 
-    Arguments
-    =========
-    form      PV form: one of 'native' (default), 'time', 'ctrl'
-    connect   whether to wait for connection (default False)
-    context   PV threading context (default None)
-    timeout   connection timeout, in seconds (default 5.0)
+def _ensure_context(func):
+    '''
+    Wrapper that ensures a method is called in the correct CA context
+
+    Assumes the instance has a `context` attribute
+
+    Raises
+    ------
+    RuntimeError
+        If the expected context (self.context) is unset (None), or the current
+        thread cannot get a valid context.  Both conditions would normally
+        result in a segmentation fault if left unchecked.
+    '''
+    @functools.wraps(func)
+    def wrapped(self, *args, **kwargs):
+        initial_context = ca.current_context()
+        expected_context = self.context
+        if expected_context is None:
+            raise RuntimeError('Expected CA context is unset')
+        elif expected_context == initial_context:
+            return func(self, *args, **kwargs)
+
+        # If not using the expected context, switch to it here:
+        if initial_context is not None:
+            ca.detach_context()
+        ca.attach_context(expected_context)
+        try:
+            return func(self, *args, **kwargs)
+        finally:
+            # Then revert back to the initial calling context
+            if initial_context is not None:
+                ca.detach_context()
+                ca.attach_context(initial_context)
+
+    return wrapped
+
+
+def get_pv(pvname, form='time', connect=False, context=None, timeout=5.0,
+           connection_callback=None, access_callback=None, callback=None,
+           verbose=False, count=None, auto_monitor=None):
+    """
+    Get a PV from PV cache or create one if needed.
+
+    Parameters
+    ---------
+    form : str, optional
+        PV form: one of 'native', 'time' (default), 'ctrl'
+    connect : bool, optional
+        whether to wait for connection (default False)
+    context : int, optional
+        PV threading context (defaults to current context)
+    timeout : float, optional
+        connection timeout, in seconds (default 5.0)
+    connection_callback : callable, optional
+        Called upon connection with keyword arguments: pvname, conn, pv
+    access_callback : callable, optional
+        Called upon update to access rights with the following signature:
+        access_callback(read_access, write_access, pv=epics.PV)
+    callback : callable, optional
+        Called upon update to change of value.  See `epics.PV.run_callback` for
+        further information regarding the signature.
+    count : int, optional
+        Number of values to request (0 or None means all available values)
+    verbose : bool, optional
+        Print additional messages relating to PV state
+    auto_monitor : bool or epics.dbr.DBE_ flags, optional
+        None: auto-monitor if count < ca.AUTOMONITOR_MAXLENGTH
+        False: do not auto-monitor
+        True: auto-monitor using ca.DEFAULT_SUBSCRIPTION_MASK
+        dbr.DBE_*: auto-monitor using this event mask. For example:
+                   `epics.dbr.DBE_ALARM|epics.dbr.DBE_LOG`
+
+    Returns
+    -------
+    pv : epics.PV
     """
 
     if form not in ('native', 'time', 'ctrl'):
         form = 'native'
 
-    thispv = None
-    if context is None:
-        context = ca.initial_context
-        if context is None:
-            context = ca.current_context()
-        if (pvname, form, context) in _PVcache_:
-            thispv = _PVcache_[(pvname, form, context)]
+    if context is not None:
+        warnings.warn(
+            'The `context` kwarg for epics.get_pv() is deprecated. New PVs '
+            'will _not_ be created in the requested context.'
+        )
+    else:
+        if ca.current_context() is None:
+            ca.use_initial_context()
+        context = ca.current_context()
+
+    pvid = (pvname, form, context)
+    thispv = _PVcache_.get(pvid, None)
 
-    start_time = time.time()
-    # not cached -- create pv (automaticall saved to cache)
     if thispv is None:
-        thispv = PV(pvname, form=form, **kws)
+        if context != ca.current_context():
+            raise RuntimeError('PV is not in cache for user-requested context')
+
+        thispv = default_pv_class(
+            pvname, form=form, callback=callback,
+            connection_callback=connection_callback,
+            access_callback=access_callback, connection_timeout=timeout,
+            count=count, verbose=verbose, auto_monitor=auto_monitor)
+
+        # Update the cache with this new instance:
+        _PVcache_[pvid] = thispv
+    else:
+        if connection_callback is not None:
+            if thispv.connected:
+                connection_callback(pvname=thispv.pvname,
+                                    conn=thispv.connected, pv=thispv)
+            thispv.connection_callbacks.append(connection_callback)
+
+        if access_callback is not None:
+            if thispv.connected:
+                access_callback(thispv.read_access, thispv.write_access,
+                                pv=thispv)
+            thispv.access_callbacks.append(access_callback)
+
+        if callback is not None:
+            idx = thispv.add_callback(callback)
+            thispv.run_callback(idx)
+
+        if auto_monitor and not thispv.auto_monitor:
+            # Start auto-monitoring, if not previously auto-monitoring:
+            thispv.auto_monitor = auto_monitor
 
     if connect:
-        thispv.wait_for_connection()
-        while not thispv.connected:
-            ca.poll()
-            if time.time()-start_time > timeout:
-                break
-        if not thispv.connected:
+        if not thispv.wait_for_connection(timeout=timeout):
             ca.write('cannot connect to %s' % pvname)
     return thispv
 
+
 def fmt_time(tstamp=None):
     "simple formatter for time values"
     if tstamp is None:
@@ -104,10 +206,12 @@ class PV(object):
         self.pvname     = pvname.strip()
         self.form       = form.lower()
         self.verbose    = verbose
-        self.auto_monitor = auto_monitor
+        self._auto_monitor = auto_monitor
         self.ftype      = None
         self.connected  = False
         self.connection_timeout = connection_timeout
+        self._user_max_count = count
+
         if self.connection_timeout is None:
             self.connection_timeout = ca.DEFAULT_CONNECTION_TIMEOUT
         self._args      = {}.fromkeys(self._fields)
@@ -127,7 +231,9 @@ class PV(object):
             self.access_callbacks = [access_callback]
 
         self.callbacks  = {}
+        self._put_complete = None
         self._monref = None  # holder of data returned from create_subscription
+        self._monref_mask = None
         self._conn_started = False
         if isinstance(callback, (tuple, list)):
             for i, thiscb in enumerate(callback):
@@ -151,27 +257,26 @@ class PV(object):
                                       use_time= self.form == 'time')
         self._args['type'] = dbr.Name(self.ftype).lower()
 
-        pvid = (self.pvname, self.form, self.context)
-        if pvid not in _PVcache_:
-            _PVcache_[pvid] = self
-
+    @_ensure_context
     def force_connect(self, pvname=None, chid=None, conn=True, **kws):
         if chid is None: chid = self.chid
-        if isinstance(chid, ctypes.c_long):
+        if hasattr(chid, 'value'):
             chid = chid.value
         self._args['chid'] = self.chid = chid
         self.__on_connect(pvname=pvname, chid=chid, conn=conn, **kws)
 
-    def force_read_access_rights(self): 
-        """force a read of access rights, not relying 
+    @_ensure_context
+    def force_read_access_rights(self):
+        """force a read of access rights, not relying
         on last event callback.
-        Note: event callback seems to fail sometimes, 
+        Note: event callback seems to fail sometimes,
         at least on initial connection on Windows 64-bit.
         """
         self._args['access'] = ca.access(self.chid)
         self._args['read_access'] = (1 == ca.read_access(self.chid))
         self._args['write_access'] = (1 == ca.write_access(self.chid))
 
+    @_ensure_context
     def __on_access_rights_event(self, read_access, write_access):
         self._args['read_access'] = read_access
         self._args['write_access'] = write_access
@@ -184,12 +289,12 @@ class PV(object):
             if callable(cb):
                 cb(read_access, write_access, pv=self)
 
+    @_ensure_context
     def __on_connect(self, pvname=None, chid=None, conn=True):
         "callback for connection events"
         # occassionally chid is still None (ie if a second PV is created
         # while __on_connect is still pending for the first one.)
         # Just return here, and connection will happen later
-        t0 = time.time()
         if self.chid is None and chid is None:
             ca.poll(5.e-4)
             return
@@ -204,37 +309,17 @@ class PV(object):
             self._args['nelm']  = count
 
             # allow reduction of elements, via count argument
-            maxcount = 0
-            if self._args['count'] is not None:
-                maxcount = self._args['count']
-                count = min(count,self._args['count'])
-
-            self._args['count']  = count
-            self._args['host']   = ca.host_name(self.chid)
-            self.ftype  = ca.promote_type(self.chid,
-                                          use_ctrl= self.form == 'ctrl',
-                                          use_time= self.form == 'time')
+            self._args['count'] = min(count, self._user_max_count or count)
+            self._args['host']  = ca.host_name(self.chid)
+            self.ftype = ca.promote_type(self.chid,
+                                         use_ctrl= self.form == 'ctrl',
+                                         use_time= self.form == 'time')
 
             _ftype_ = dbr.Name(self.ftype).lower()
             self._args['type'] = _ftype_
             self._args['typefull'] = _ftype_
             self._args['ftype'] = dbr.Name(_ftype_, reverse=True)
-
-            if self.auto_monitor is None:
-                self.auto_monitor = count < ca.AUTOMONITOR_MAXLENGTH
-            if self._monref is None and self.auto_monitor:
-                # you can explicitly request a subscription mask
-                # (ie dbr.DBE_ALARM|dbr.DBE_LOG) by passing it as the
-                # auto_monitor arg, otherwise if you specify 'True' you'll
-                # just get the default set in ca.DEFAULT_SUBSCRIPTION_MASK
-                mask = None
-                if isinstance(self.auto_monitor, int):
-                    mask = self.auto_monitor
-                self._monref = ca.create_subscription(self.chid,
-                                         use_ctrl=(self.form == 'ctrl'),
-                                         use_time=(self.form == 'time'),
-                                         callback=self.__on_changes,
-                                                      mask=mask, count=maxcount)
+            self._check_auto_monitor()
 
         for conn_cb in self.connection_callbacks:
             if hasattr(conn_cb, '__call__'):
@@ -249,18 +334,92 @@ class PV(object):
         # threads from thinking a connection is complete when it is actually
         # still in progress.
         self.connected = conn
-        return
 
+    @_ensure_context
+    def _clear_auto_monitor_subscription(self):
+        'Clear an auto-monitor subscription, if set'
+        if self._monref is None:
+            return
+
+        cback, uarg, evid = self._monref
+
+        self._monref = None
+        self._monref_mask = None
+        ca.clear_subscription(evid)
+
+    @_ensure_context
+    def _check_auto_monitor(self):
+        '''
+        Check the auto-monitor status
+
+        Clears or adds monitor, if necessary.
+        '''
+        count = self.count
+        chid = self.chid
+
+        if count is None or chid is None:
+            return
+
+        if self._auto_monitor is None:
+            self._auto_monitor = count < ca.AUTOMONITOR_MAXLENGTH
+
+        if not self._auto_monitor:
+            # Turn off auto-monitoring, if necessary:
+            return self._clear_auto_monitor_subscription()
+
+        mask = (ca.DEFAULT_SUBSCRIPTION_MASK
+                if self._auto_monitor is True
+                else self._auto_monitor)
+
+        if self._monref is not None:
+            if self._monref_mask == mask:
+                # Same mask; no need to redo subscription
+                return
+
+            # New mask.
+            self._clear_auto_monitor_subscription()
+
+        self._monref_mask = mask
+        self._monref = ca.create_subscription(
+            self.chid,
+            use_ctrl=(self.form == 'ctrl'),
+            use_time=(self.form == 'time'),
+            callback=self.__on_changes,
+            mask=mask,
+            count=self._user_max_count or 0
+        )
+
+    @property
+    def auto_monitor(self):
+        '''
+        Whether auto_monitor is enabled or not. May be one of the following::
+
+            None: auto-monitor if count < ca.AUTOMONITOR_MAXLENGTH
+            False: do not auto-monitor
+            True: auto-monitor using ca.DEFAULT_SUBSCRIPTION_MASK
+            dbr.DBE_*: auto-monitor using this event mask. For example:
+                       `epics.dbr.DBE_ALARM|epics.dbr.DBE_LOG`
+        '''
+        return self._auto_monitor
+
+    @auto_monitor.setter
+    @_ensure_context
+    def auto_monitor(self, value):
+        self._auto_monitor = value
+        self._check_auto_monitor()
+
+    @property
+    def auto_monitor_mask(self):
+        'The current mask in use for auto-monitoring'
+        return self._monref_mask
+
+    @_ensure_context
     def wait_for_connection(self, timeout=None):
         """wait for a connection that started with connect() to finish"""
-
-        # make sure we're in the CA context used to create this PV
-        if self.context != ca.current_context():
-            ca.attach_context(self.context)
         if not self.connected:
             start_time = time.time()
             if not self._conn_started:
-                self.connect()
+                self.connect(timeout=timeout)
 
             if not self.connected:
                 if timeout is None:
@@ -269,6 +428,7 @@ class PV(object):
                     ca.poll()
         return self.connected
 
+    @_ensure_context
     def connect(self, timeout=None):
         "check that a PV is connected, forcing a connection if needed"
         if not self.connected:
@@ -278,23 +438,20 @@ class PV(object):
         self._conn_started = True
         return self.connected and self.ftype is not None
 
+    @_ensure_context
     def clear_auto_monitor(self):
-        """turn off auto-monitoring: must reconnect to re-enable monitoring"""
+        """turn off auto-monitoring"""
         self.auto_monitor = False
-        if self._monref is not None:
-            evid = self._monref[2]
-            ca.clear_subscription(evid)
-            self._monref = None
 
     def reconnect(self):
         "try to reconnect PV"
-        self.auto_monitor = None
-        self._monref = None
+        self._clear_auto_monitor_subscription()
         self.connected = False
         self._conn_started = False
         self.force_connect()
         return self.wait_for_connection()
 
+    @_ensure_context
     def poll(self, evt=1.e-4, iot=1.0):
         "poll for changes"
         ca.poll(evt=evt, iot=iot)
@@ -313,60 +470,152 @@ class PV(object):
                     monitor callback (True, default) or to make an
                     explicit CA call for the value.
 
-        >>> p.get('13BMD:m1.DIR')
+        >>> get_pv('13BMD:m1.DIR').get()
         0
-        >>> p.get('13BMD:m1.DIR', as_string=True)
+        >>> get_pv('13BMD:m1.DIR').get(as_string=True)
         'Pos'
+
+        If the Channel Access status code sent by the IOC indicates a failure,
+        this method will raise the exception ChannelAccessGetFailure.
+        """
+        data = self.get_with_metadata(count=count, as_string=as_string,
+                                      as_numpy=as_numpy, timeout=timeout,
+                                      with_ctrlvars=with_ctrlvars,
+                                      use_monitor=use_monitor)
+        return (data['value']
+                if data is not None
+                else None)
+
+    @_ensure_context
+    def get_with_metadata(self, count=None, as_string=False, as_numpy=True,
+                          timeout=None, with_ctrlvars=False, form=None,
+                          use_monitor=True, as_namespace=False):
+        """Returns a dictionary of the current value and associated metadata
+
+        count         explicitly limit count for array data
+        as_string     flag(True/False) to get a string representation
+                      of the value.
+        as_numpy      flag(True/False) to use numpy array as the
+                      return type for array data.
+        timeout       maximum time to wait for value to be received.
+                      (default = 0.5 + log10(count) seconds)
+        use_monitor   flag(True/False) to use value from latest
+                      monitor callback (True, default) or to make an
+                      explicit CA call for the value.
+        form          {'time', 'ctrl', None} optionally change the type of the
+                      get request
+        as_namespace  Change the return type to that of a namespace with
+                      support for tab-completion
+
+        >>> get_pv('13BMD:m1.DIR', form='time').get_with_metadata()
+        {'value': 0, 'status': 0, 'severity': 0}
+        >>> get_pv('13BMD:m1.DIR').get_with_metadata(form='ctrl')
+        {'value': 0, 'lower_ctrl_limit': 0, ...}
+        >>> get_pv('13BMD:m1.DIR').get_with_metadata(as_string=True)
+        {'value': 'Pos', 'status': 0, 'severity': 0}
+        >>> ns = get_pv('13BMD:m1.DIR').get_with_metadata(as_string=True,
+                                                          as_namespace=True)
+        >>> ns
+        namespace(value='Pos', status=0, severity=0, ...)
+        >>> ns.status
+        0
         """
         if not self.wait_for_connection(timeout=timeout):
             return None
+
+        if form is None:
+            form = self.form
+            ftype = self.ftype
+        else:
+            ftype = ca.promote_type(self.chid,
+                                    use_ctrl=(form == 'ctrl'),
+                                    use_time=(form == 'time'))
+
         if with_ctrlvars and getattr(self, 'units', None) is None:
-            self.get_ctrlvars()
+            if form != 'ctrl':
+                # ctrlvars will be updated as the get completes, since this
+                # metadata comes bundled with our DBR_CTRL* request.
+                pass
+            else:
+                self.get_ctrlvars()
+
+        try:
+            cached_length = len(self._args['value'])
+        except TypeError:
+            cached_length = 1
 
         if ((not use_monitor) or
-            (not self.auto_monitor) or
-            (self._args['value'] is None) or
-            (count is not None and count > len(self._args['value']))):
+                (not self.auto_monitor) or
+                (ftype != self.ftype) or
+                (self._args['value'] is None) or
+                (count is not None and count > cached_length)):
 
             # respect count argument on subscription also for calls to get
             if count is None and self._args['count']!=self._args['nelm']:
                 count = self._args['count']
-            ca_get = ca.get
-            if ca.get_cache(self.pvname)['value'] is not None:
-                ca_get = ca.get_complete
 
-            self._args['value'] = ca_get(self.chid, ftype=self.ftype,
-                                         count=count, timeout=timeout,
-                                         as_numpy=as_numpy)
-        val = self._args['value']
+            # ca.get_with_metadata will handle multiple requests for the same
+            # PV internally, so there is no need to change between
+            # `get_with_metadata` and `get_complete_with_metadata` here.
+            md = ca.get_with_metadata(
+                self.chid, ftype=ftype, count=count, timeout=timeout,
+                as_numpy=as_numpy)
+            if md is None:
+                # Get failed. Indicate with a `None` as the return value
+                return
+
+            # Update value and all included metadata. Depending on the PV
+            # form, this could include timestamp, alarm information,
+            # ctrlvars, and so on.
+            self._args.update(**md)
+
+            if with_ctrlvars and form != 'ctrl':
+                # If the user requested ctrlvars and they were not included in
+                # the request, return all metadata.
+                md = self._args.copy()
+
+            val = md['value']
+        else:
+            md = self._args.copy()
+            val = self._args['value']
+
         if as_string:
-            return self._set_charval(val)
-        if self.count <= 1 or val is None:
-            return val
+            char_value = self._set_charval(val, force_long_string=as_string)
+            md['value'] = char_value
+        elif self.nelm <= 1 or val is None:
+            pass
+        else:
+            # After this point:
+            #   * self.nelm is > 1
+            #   * val should be set and a sequence
+            try:
+                len(val)
+            except TypeError:
+                # Edge case where a scalar value leaks through ca.unpack()
+                val = [val]
 
-        # After this point:
-        #   * self.count is > 1
-        #   * val should be set and a sequence
-        try:
-            len(val)
-        except TypeError:
-            # Edge case where a scalar value leaks through ca.unpack()
-            val = [val]
+            if count is None:
+                count = len(val)
 
-        if count is None:
-            count = len(val)
+            if (as_numpy and ca.HAS_NUMPY and
+                    not isinstance(val, ca.numpy.ndarray)):
+                val = ca.numpy.asarray(val)
+            elif (not as_numpy and ca.HAS_NUMPY and
+                    isinstance(val, ca.numpy.ndarray)):
+                val = val.tolist()
 
-        if (as_numpy and ca.HAS_NUMPY and
-                not isinstance(val, ca.numpy.ndarray)):
-            val = ca.numpy.asarray(val)
-        elif (not as_numpy and ca.HAS_NUMPY and
-                isinstance(val, ca.numpy.ndarray)):
-            val = val.tolist()
-        # allow asking for less data than actually exists in the cached value
-        if count < len(val):
-            val = val[:count]
-        return val
+            # allow asking for less data than actually exists in the cached value
+            if count < len(val):
+                val = val[:count]
 
+            # Update based on the requested type:
+            md['value'] = val
+
+        if as_namespace:
+            return Namespace(**md)
+        return md
+
+    @_ensure_context
     def put(self, value, wait=False, timeout=30.0,
             use_complete=False, callback=None, callback_data=None):
         """set value for PV, optionally waiting until the processing is
@@ -387,18 +636,22 @@ class PV(object):
                     if val == value:
                         value = ival
                         break
-        if use_complete and callback is None:
-            callback = self.__putCallbackStub
+
+        def _put_callback(pvname=None, **kws):
+            self._put_complete = True
+            if callback is not None:
+                callback(pvname=pvname, **kws)
+
+        self._put_complete = (False
+                              if use_complete
+                              else None)
+
         return ca.put(self.chid, value,
                       wait=wait, timeout=timeout,
-                      callback=callback,
+                      callback=_put_callback if use_complete or callback else None,
                       callback_data=callback_data)
 
-    def __putCallbackStub(self, pvname=None, **kws):
-        "null put-calback, so that the put_complete attribute is valid"
-        pass
-
-    def _set_charval(self, val, call_ca=True):
+    def _set_charval(self, val, call_ca=True, force_long_string=False):
         """ sets the character representation of the value.
         intended only for internal use"""
         if val is None:
@@ -410,7 +663,8 @@ class PV(object):
             self._args['char_value'] = val
             return val
         # char waveform as string
-        if ntype == dbr.CHAR and self.count < ca.AUTOMONITOR_MAXLENGTH:
+        if ntype == dbr.CHAR and (self.count < ca.AUTOMONITOR_MAXLENGTH or
+                force_long_string is True):
             if ca.HAS_NUMPY and isinstance(val, ca.numpy.ndarray):
                 # a numpy array
                 val = val.tolist()
@@ -469,21 +723,25 @@ class PV(object):
         self._args['char_value'] = cval
         return cval
 
+    @_ensure_context
     def get_ctrlvars(self, timeout=5, warn=True):
         "get control values for variable"
         if not self.wait_for_connection():
             return None
         kwds = ca.get_ctrlvars(self.chid, timeout=timeout, warn=warn)
-        self._args.update(kwds)
+        if kwds is not None:
+            self._args.update(kwds)
         self.force_read_access_rights()
         return kwds
 
+    @_ensure_context
     def get_timevars(self, timeout=5, warn=True):
         "get time values for variable"
         if not self.wait_for_connection():
             return None
         kwds = ca.get_timevars(self.chid, timeout=timeout, warn=warn)
-        self._args.update(kwds)
+        if kwds is not None:
+            self._args.update(kwds)
         return kwds
 
 
@@ -504,6 +762,7 @@ class PV(object):
                                      self._args['char_value'], now))
         self.run_callbacks()
 
+    @_ensure_context
     def run_callbacks(self):
         """run all user-defined callbacks with the current data
 
@@ -514,6 +773,7 @@ class PV(object):
         for index in sorted(list(self.callbacks.keys())):
             self.run_callback(index)
 
+    @_ensure_context
     def run_callback(self, index):
         """run a specific user-defined callback, specified by index,
         with the current data
@@ -560,6 +820,7 @@ class PV(object):
                 self.run_callback(index)
         return index
 
+    @_ensure_context
     def remove_callback(self, index=None):
         """remove a callback by index"""
         if index in self.callbacks:
@@ -568,7 +829,7 @@ class PV(object):
 
     def clear_callbacks(self):
         "clear all callbacks"
-        self.callbacks = {}
+        self.callbacks.clear()
 
     def _getinfo(self):
         "get information paragraph"
@@ -695,12 +956,13 @@ class PV(object):
             return self._getarg('count')
 
     @property
+    @_ensure_context
     def nelm(self):
         """native count (number of elements).
         For array data this will return the full array size (ie, the
         .NELM field).  See also 'count' property"""
-        if self._getarg('count') == 1:
-            return 1
+        # if self._getarg('count') == 1:
+        #    return 1
         return ca.element_count(self.chid)
 
     @property
@@ -801,11 +1063,8 @@ class PV(object):
 
     @property
     def put_complete(self):
-        "returns True if a put-with-wait has completed"
-        putdone_data = ca._put_done.get(self.pvname, None)
-        if putdone_data is not None:
-            return putdone_data[0]
-        return True
+        "returns True if the last put-with-wait has completed"
+        return self._put_complete
 
     def __repr__(self):
         "string representation"
@@ -825,6 +1084,7 @@ class PV(object):
         except AttributeError:
             return False
 
+    @_ensure_context
     def disconnect(self):
         "disconnect PV"
         self.connected = False
@@ -834,24 +1094,20 @@ class PV(object):
         if pvid in _PVcache_:
             _PVcache_.pop(pvid)
 
-        if self._monref is not None:
-            cback, uarg, evid = self._monref
-            ctx = ca.current_context()
-            if self.pvname in ca._cache[ctx]:
+        cache_item = ca._cache[ctx].pop(self.pvname, None)
+        if cache_item is not None:
+            if self._monref is not None:
                 # atexit may have already cleared the subscription
-                ca.clear_subscription(evid)
-                ca._cache[ctx].pop(self.pvname)
-            del cback
-            del uarg
-            del evid
-            try:
-                self._monref = None
-                self._args   = {}.fromkeys(self._fields)
-            except:
-                pass
+                self._clear_auto_monitor_subscription()
 
+            # TODO: clear channel should be called as well
+            # ca.clear_channel(cache_item.chid)
+
+        self._monref = None
+        self._monref_mask = None
+        self.clear_callbacks()
+        self._args = {}.fromkeys(self._fields)
         ca.poll(evt=1.e-3, iot=1.0)
-        self.callbacks = {}
 
     def __del__(self):
         if getattr(ca, 'libca', None) is None:
@@ -861,3 +1117,7 @@ class PV(object):
             self.disconnect()
         except:
             pass
+
+
+# Allow advanced users to customize the class of PV that `get_pv` would return:
+default_pv_class = PV
